@@ -1,42 +1,56 @@
 """Base classes and function for REST, WMS, and WMF services."""
+from __future__ import annotations
+
 import contextlib
-import logging
 import os
 import sys
 import urllib.parse as urlparse
 import uuid
+from dataclasses import dataclass
 from pathlib import Path
-from typing import Any, Dict, Iterator, List, Mapping, Optional, Sequence, Tuple, Union
+from typing import Any, Iterator, Mapping, Sequence, Union
 
 import async_retriever as ar
 import cytoolz as tlz
 import pyproj
+from loguru import logger
 from owslib.wfs import WebFeatureService
 from owslib.wms import WebMapService
-from pydantic import BaseModel, validator
 from shapely.geometry import LineString, MultiPoint, Point, Polygon
 
 from . import utils
 from .exceptions import (
-    InvalidInputType,
-    InvalidInputValue,
-    MissingInputs,
+    InputTypeError,
+    InputValueError,
+    MissingInputError,
     ServiceError,
-    ServiceUnavailable,
-    ZeroMatched,
+    ServiceUnavailableError,
+    ZeroMatchedError,
 )
 
-logger = logging.getLogger(__name__)
-logger.setLevel(logging.INFO)
-handler = logging.StreamHandler(sys.stdout)
-handler.setFormatter(logging.Formatter(""))
-logger.handlers = [handler]
-logger.propagate = False
-DEF_CRS = "epsg:4326"
-EXPIRE = -1
+logger.configure(
+    handlers=[
+        {
+            "sink": sys.stdout,
+            "colorize": True,
+            "format": " | ".join(
+                [
+                    "{time:YYYY-MM-DD at HH:mm:ss}",  # noqa: FS003
+                    "{name: ^15}.{function: ^15}:{line: >3}",  # noqa: FS003
+                    "{message}",  # noqa: FS003
+                ]
+            ),
+        }
+    ]
+)
+if os.environ.get("HYRIVER_VERBOSE", "false").lower() == "true":
+    logger.enable("pygeoogc")
+else:
+    logger.disable("pygeoogc")
+CRSTYPE = Union[int, str, pyproj.CRS]
 
 
-def validate_version(val: str, valid_versions: List[str]) -> str:
+def validate_version(val: str, valid_versions: list[str]) -> str:
     """Validate version from a list of valid versions.
 
     Parameters
@@ -52,91 +66,23 @@ def validate_version(val: str, valid_versions: List[str]) -> str:
         Validated version value.
     """
     if val not in valid_versions:
-        raise InvalidInputValue("version", valid_versions)
+        raise InputValueError("version", valid_versions)
     return val
 
 
-class RESTValidator(BaseModel):
-    """Validate ArcGISRESTful inputs.
-
-    Parameters
-    ----------
-    base_url : str, optional
-        The ArcGIS RESTful service url. The URL must either include a layer number
-        after the last ``/`` in the url or the target layer must be passed as an argument.
-    layer : int, optional
-        Target layer number, defaults to None. If None layer number must be included as after
-        the last ``/`` in ``base_url``.
-    outformat : str, optional
-        One of the output formats offered by the selected layer. If not correct
-        a list of available formats is shown, defaults to ``geojson``.
-    outfields : str or list
-        The output fields to be requested. Setting ``*`` as outfields requests
-        all the available fields which is the default setting.
-    crs : str, optional
-        The spatial reference of the output data, defaults to EPSG:4326
-    max_workers : int, optional
-        Max number of simultaneous requests, default to 2. Note
-        that some services might face issues when several requests are sent
-        simultaneously and will return the requests partially. It's recommended
-        to avoid using too many workers unless you are certain the web service
-        can handle it.
-    verbose : bool, optional
-        If True, prints information about the requests and responses,
-        defaults to False.
-    disable_retry : bool, optional
-        If ``True`` in case there are any failed queries, no retrying attempts
-        is done and object IDs of the failed requests is saved to a text file
-        which its path can be accessed via ``self.failed_path``.
-    """
-
-    base_url: str
-    layer: Optional[int] = None
-    outformat: str = "geojson"
-    outfields: Union[List[str], str] = "*"
-    crs: str = DEF_CRS
-    max_workers: int = 1
-    verbose: bool = False
-    disable_retry: bool = False
-
-    @staticmethod
-    def _split_url(url: str) -> Tuple[str, Optional[int]]:
-        """Check if layer is included in url, if so separate and return them."""
+def split_url(url: str, layer: int | None) -> tuple[str, int]:
+    """Check if layer is included in url, if so separate and return them."""
+    url = urlparse.unquote(url)
+    if layer is None:
         url_split = urlparse.urlsplit(url.rstrip("/"))
         url_path, lyr = url_split.path.rsplit("/", 1)
         url_split = url_split._replace(path=url_path)
         try:
             return urlparse.urlunsplit(url_split), int(lyr)
-        except ValueError:
-            return url, None
-
-    @validator("base_url")
-    def _layer_from_url(cls, v: str) -> Tuple[str, Optional[int]]:
-        return cls._split_url(urlparse.unquote(v))
-
-    @validator("layer")
-    def _integer_layer(cls, v: Optional[int], values: Dict[str, Any]) -> int:
-        if v is None:
-            try:
-                return int(values["base_url"][1])
-            except TypeError as ex:
-                msg = "Either layer must be passed as an argument or be included in ``base_url``"
-                raise MissingInputs(msg) from ex
-        return v
-
-    @validator("outfields")
-    def _outfields_to_list(cls, v: Union[List[str], str]) -> List[str]:
-        return v if isinstance(v, list) else [v]
-
-    @validator("crs")
-    def _valid_crs(cls, v: str) -> str:
-        return utils.validate_crs(v)
-
-    @validator("max_workers")
-    def _positive_integer_threads(cls, v: int) -> int:
-        if v <= 0:
-            raise InvalidInputType("max_workers", "positive integer")
-        return v
+        except ValueError as ex:
+            msg = "Either layer must be passed as an argument or be included in ``base_url``"
+            raise MissingInputError(msg) from ex
+    return url, layer
 
 
 class ArcGISRESTfulBase:
@@ -158,8 +104,8 @@ class ArcGISRESTfulBase:
     outfields : str or list
         The output fields to be requested. Setting ``*`` as outfields requests
         all the available fields which is the default setting.
-    crs : str, optional
-        The spatial reference of the output data, defaults to EPSG:4326
+    crs : str, int, or pyproj.CRS, optional
+        The spatial reference of the output data, defaults to ``epsg:4326``
     max_workers : int, optional
         Max number of simultaneous requests, default to 2. Note
         that some services might face issues when several requests are sent
@@ -172,57 +118,48 @@ class ArcGISRESTfulBase:
     disable_retry : bool, optional
         If ``True`` in case there are any failed queries, no retrying attempts
         is done and object IDs of the failed requests is saved to a text file
-        which its path can be accessed via ``self.failed_path``.
+        which its ipath can be accessed via ``self.failed_path``.
     """
 
     def __init__(
         self,
         base_url: str,
-        layer: Optional[int] = None,
+        layer: int | None = None,
         outformat: str = "geojson",
-        outfields: Union[List[str], str] = "*",
-        crs: str = DEF_CRS,
+        outfields: list[str] | str = "*",
+        crs: CRSTYPE = 4326,
         max_workers: int = 1,
         verbose: bool = False,
         disable_retry: bool = False,
     ) -> None:
-        validated = RESTValidator(
-            base_url=base_url,
-            layer=layer,
-            outformat=outformat,
-            outfields=outfields,
-            crs=crs,
-            max_workers=max_workers,
-            verbose=verbose,
-            disable_retry=disable_retry,
-        )
-        self.base_url = validated.base_url[0]
-        self.layer = validated.layer
-        self.outformat = validated.outformat
-        self.outfields = validated.outfields
-        self.crs = validated.crs
-        self.max_workers = validated.max_workers
-        self.verbose = validated.verbose
-        self.disable_retry = validated.disable_retry
+        self.base_url, self.layer = split_url(base_url, layer)
+        self.outformat = outformat
+        self.outfields = outfields if isinstance(outfields, (list, tuple)) else [outfields]
+        self.crs = utils.validate_crs(crs)
+        self.max_workers = max_workers
+        if self.max_workers < 1 or not isinstance(self.max_workers, int):
+            raise InputTypeError("max_workers", "positive integer > 1")
+        self.verbose = verbose
+        self.disable_retry = disable_retry
 
         self.out_sr = pyproj.CRS(self.crs).to_epsg()
         self.n_features = 0
         self.url = f"{self.base_url}/{self.layer}"
         self.query_url = f"{self.url}/query"
-        self.valid_layers: Dict[str, str] = {}
-        self.query_formats: List[str] = []
-        self.extent: Optional[Tuple[float, float, float, float]] = None
-        self.units: Optional[str] = None
+        self.valid_layers: dict[str, str] = {}
+        self.query_formats: list[str] = []
+        self.extent: tuple[float, float, float, float] | None = None
+        self.units: str | None = None
         self.max_nrecords: int = 1000
-        self.valid_fields: List[str] = []
-        self.field_types: Dict[str, str] = {}
-        self.feature_types: Optional[Dict[int, str]] = None
+        self.valid_fields: list[str] = []
+        self.field_types: dict[str, str] = {}
+        self.feature_types: dict[int, str] | None = None
         self.return_m: bool = False
         self.return_geom: bool = True
         self.n_missing: int = 0
         self.total_n_features: int = 0
-        self.failed_path: Union[str, Path] = ""
-        self.request_id: Optional[str] = None
+        self.failed_path: str | Path = ""
+        self.request_id: str | None = None
 
         self.initialize_service()
 
@@ -241,9 +178,9 @@ class ArcGISRESTfulBase:
         while True:
             try:
                 crs = utils.validate_crs(extent["spatialReference"].get(next(crs_iter)))
-                self.extent = utils.match_crs(bounds, crs, DEF_CRS)
+                self.extent = utils.match_crs(bounds, crs, 4326)
                 break
-            except InvalidInputType:
+            except InputTypeError:
                 continue
             except StopIteration as ex:
                 raise ServiceError(self.base_url) from ex
@@ -276,21 +213,21 @@ class ArcGISRESTfulBase:
         self._set_service_properties()
         if f"{self.layer}" not in self.valid_layers:
             valids = [f'"{i}" for {n}' for i, n in self.valid_layers.items()]
-            raise InvalidInputValue("layer", valids)
+            raise InputValueError("layer", valids)
 
         self._set_layer_properties()
 
         if self.outformat.lower() not in self.query_formats:
-            raise InvalidInputValue("outformat", self.query_formats)
+            raise InputValueError("outformat", self.query_formats)
 
         if any(f not in self.valid_fields for f in self.outfields):
-            raise InvalidInputValue("outfields", self.valid_fields)
+            raise InputValueError("outfields", self.valid_fields)
 
-    def partition_oids(self, oids: Union[List[int], int]) -> Iterator[Tuple[str, ...]]:
+    def partition_oids(self, oids: list[int] | int) -> Iterator[tuple[str, ...]]:
         """Partition feature IDs based on ``self.max_nrecords``."""
         oid_list = [oids] if isinstance(oids, int) else set(oids)
         if len(oid_list) == 0:
-            raise ZeroMatched
+            raise ZeroMatchedError
 
         self.n_features = len(oid_list)
         if not self.disable_retry and self.verbose:
@@ -298,8 +235,8 @@ class ArcGISRESTfulBase:
         return tlz.partition_all(self.max_nrecords, [str(i) for i in oid_list])  # type: ignore
 
     def _cleanup_resp(
-        self, resp: List[Dict[str, Any]], payloads: Sequence[Dict[str, str]]
-    ) -> List[Dict[str, Any]]:
+        self, resp: list[dict[str, Any]], payloads: Sequence[dict[str, str]]
+    ) -> list[dict[str, Any]]:
         """Remove failed responses."""
         fails = [i for i, r in enumerate(resp) if "error" in r]
 
@@ -342,10 +279,10 @@ class ArcGISRESTfulBase:
 
     def get_features(
         self,
-        featureids: Iterator[Tuple[str, ...]],
+        featureids: Iterator[tuple[str, ...]],
         return_m: bool = False,
         return_geom: bool = True,
-    ) -> List[Dict[str, Any]]:
+    ) -> list[dict[str, Any]]:
         """Get features based on the feature IDs.
 
         Parameters
@@ -381,19 +318,13 @@ class ArcGISRESTfulBase:
 
         resp = self._cleanup_resp(resp, payloads)
         if len(resp) == 0:
-            raise ZeroMatched
+            raise ZeroMatchedError
         return resp
 
     def esri_query(
         self,
-        geom: Union[
-            LineString,
-            Polygon,
-            Point,
-            MultiPoint,
-            Tuple[float, float, float, float],
-        ],
-        geo_crs: Union[str, pyproj.CRS] = DEF_CRS,
+        geom: (LineString | Polygon | Point | MultiPoint | tuple[float, float, float, float]),
+        geo_crs: CRSTYPE = 4326,
     ) -> Mapping[str, str]:
         """Generate geometry queries based on ESRI template."""
         geom = utils.match_crs(geom, geo_crs, self.crs)
@@ -413,11 +344,11 @@ class ArcGISRESTfulBase:
         if isinstance(geom, LineString):
             return utils.ESRIGeomQuery(geom, self.out_sr).polyline()
 
-        raise InvalidInputType("geom", "LineString, Polygon, Point, MultiPoint, tuple")
+        raise InputTypeError("geom", "LineString, Polygon, Point, MultiPoint, tuple")
 
     def _retry(
         self, return_m: bool, return_geo: bool, partition_fac: float
-    ) -> List[Dict[str, Any]]:
+    ) -> list[dict[str, Any]]:
         """Retry failed requests."""
         with open(self.failed_path) as f:
             oids = [int(i) for i in f.read().splitlines()]
@@ -429,7 +360,7 @@ class ArcGISRESTfulBase:
 
         return features
 
-    def retry_failed_requests(self) -> List[Dict[str, Any]]:
+    def retry_failed_requests(self) -> list[dict[str, Any]]:
         """Retry failed requests."""
         retry = self.disable_retry
         self.disable_retry = True
@@ -451,8 +382,8 @@ class ArcGISRESTfulBase:
         return list(tlz.concat(features))
 
     def get_response(
-        self, url: str, payloads: Sequence[Dict[str, str]], method: str = "GET"
-    ) -> List[Dict[str, Any]]:
+        self, url: str, payloads: Sequence[dict[str, str]], method: str = "GET"
+    ) -> list[dict[str, Any]]:
         """Send payload and get the response."""
         req_key = "params" if method == "GET" else "data"
         try:
@@ -463,7 +394,7 @@ class ArcGISRESTfulBase:
                 max_workers=self.max_workers,
             )
         except ValueError:
-            raise ZeroMatched
+            raise ZeroMatchedError
 
     def __repr__(self) -> str:
         """Print the service configuration."""
@@ -483,67 +414,81 @@ class ArcGISRESTfulBase:
         )
 
 
-class WMSBase(BaseModel):
+@dataclass
+class WMSBase:
     """Base class for accessing a WMS service.
 
     Parameters
     ----------
     url : str
         The base url for the WMS service e.g., https://www.mrlc.gov/geoserver/mrlc_download/wms
-    layers : str or list
+    layers : str or list, optional
         A layer or a list of layers from the service to be downloaded. You can pass an empty
         string to get a list of available layers.
-    outformat : str
+    outformat : str, optional
         The data format to request for data from the service. You can pass an empty
         string to get a list of available output formats.
     version : str, optional
         The WMS service version which should be either 1.1.1 or 1.3.0, defaults to 1.3.0.
-    crs : str, optional
+    crs : str, int, or pyproj.CRS, optional
         The spatial reference system to be used for requesting the data, defaults to
         ``epsg:4326``.
+    validation : bool, optional
+        Validate the input arguments from the WMS service, defaults to True. Set this
+        to False if you are sure all the WMS settings such as layer and crs are correct
+        to avoid sending extra requests.
     """
 
     url: str
-    layers: Union[str, List[str]]
-    outformat: str
+    layers: str | list[str] = ""
+    outformat: str = ""
     version: str = "1.3.0"
-    crs: str = DEF_CRS
+    crs: CRSTYPE = 4326
+    validation: bool = True
 
-    @validator("crs")
-    def _valid_crs(cls, v: str) -> str:
-        return utils.validate_crs(v)
+    def __post_init__(self) -> None:
+        """Validate crs."""
+        self.crs_str = utils.validate_crs(self.crs)
+        self.version = validate_version(self.version, ["1.1.1", "1.3.0"])
+        self.get_service_options()
+        if self.validation:
+            self.validate_wms()
 
-    @validator("version")
-    def _version(cls, v: str) -> str:
-        return validate_version(v, ["1.1.1", "1.3.0"])
-
-    def validate_wms(self) -> None:
+    def get_service_options(self) -> None:
         """Validate input arguments with the WMS service."""
         try:
             wms = WebMapService(self.url, version=self.version)
         except AttributeError as ex:
-            raise ServiceUnavailable(self.url) from ex
+            raise ServiceUnavailableError(self.url) from ex
 
+        self.available_layer = {wms[lyr].name: wms[lyr].title for lyr in list(wms.contents)}
+        self.available_outformat = wms.getOperationByName("GetMap").formatOptions
+        self.available_outformat = [f.lower() for f in self.available_outformat]
+        self.available_crs = {
+            lyr: [s.lower() for s in wms[lyr].crsOptions] for lyr in self.available_layer
+        }
+
+    def validate_wms(self) -> None:
+        """Validate input arguments with the WMS service."""
         layers = [self.layers] if isinstance(self.layers, str) else self.layers
-        valid_layers = {wms[lyr].name: wms[lyr].title for lyr in list(wms.contents)}
-        if any(lyr not in valid_layers.keys() for lyr in layers):
-            raise InvalidInputValue("layers", (f"{n} for {t}" for n, t in valid_layers.items()))
+        if any(lyr not in self.available_layer.keys() for lyr in layers):
+            raise InputValueError(
+                "layers", (f"{n} for {t}" for n, t in self.available_layer.items())
+            )
 
-        valid_outformats = wms.getOperationByName("GetMap").formatOptions
-        if self.outformat not in valid_outformats:
-            raise InvalidInputValue("outformat", valid_outformats)
+        if self.outformat not in self.available_outformat:
+            raise InputValueError("outformat", self.available_outformat)
 
-        valid_crss = {lyr: [s.lower() for s in wms[lyr].crsOptions] for lyr in layers}
-        if any(self.crs.lower() not in valid_crss[lyr] for lyr in layers):
-            _valid_crss = (f"{lyr}: {', '.join(cs)}\n" for lyr, cs in valid_crss.items())
-            raise InvalidInputValue("CRS", _valid_crss)
+        if any(self.crs_str.lower() not in self.available_crs[lyr] for lyr in layers):
+            _valid_crss = (f"{lyr}: {', '.join(cs)}\n" for lyr, cs in self.available_crs.items())
+            raise InputValueError("CRS", _valid_crss)
 
-    def get_validlayers(self) -> Dict[str, str]:
+    def get_validlayers(self) -> dict[str, str]:
         """Get the layers supported by the WMS service."""
         try:
             wms = WebMapService(self.url, version=self.version)
         except AttributeError as ex:
-            raise ServiceUnavailable(self.url) from ex
+            raise ServiceUnavailableError(self.url) from ex
 
         return {wms[lyr].name: wms[lyr].title for lyr in list(wms.contents)}
 
@@ -556,11 +501,12 @@ class WMSBase(BaseModel):
             + f"Version: {self.version}\n"
             + f"Layers: {', '.join(lyr for lyr in layers)}\n"
             + f"Output Format: {self.outformat}\n"
-            + f"Output CRS: {self.crs}"
+            + f"Output CRS: {self.crs_str}"
         )
 
 
-class WFSBase(BaseModel):
+@dataclass
+class WFSBase:
     """Base class for WFS service.
 
     Parameters
@@ -577,7 +523,7 @@ class WFSBase(BaseModel):
     version : str, optional
         The WFS service version which should be either ``1.0.0``, ``1.1.0``, or
         ``2.0.0``. Defaults to ``2.0.0``.
-    crs: str, optional
+    crs : str, int, or pyproj.CRS, optional
         The spatial reference system to be used for requesting the data, defaults to
         ``epsg:4326``.
     read_method : str, optional
@@ -587,70 +533,68 @@ class WFSBase(BaseModel):
         The maximum number of records in a single request to be retrieved from the service,
         defaults to 1000. If the number of requested records is greater than this value,
         the query will be split into multiple requests.
+    validation : bool, optional
+        Validate the input arguments from the WFS service, defaults to True. Set this
+        to False if you are sure all the WFS settings such as layer and crs are correct
+        to avoid sending extra requests.
     """
 
     url: str
-    layer: Optional[str] = None
-    outformat: Optional[str] = None
+    layer: str | None = None
+    outformat: str | None = None
     version: str = "2.0.0"
-    crs: str = DEF_CRS
+    crs: CRSTYPE = 4326
     read_method: str = "json"
     max_nrecords: int = 1000
+    validation: bool = True
 
-    @validator("read_method")
-    def _read_method(cls, v: str) -> str:
+    def __post_init__(self) -> None:
+        """Validate crs."""
+        self.crs_str = utils.validate_crs(self.crs)
+        self.version = validate_version(self.version, ["1.0.0", "1.1.0", "2.0.0"])
         valid_methods = ["json", "binary", "text"]
-        if v not in valid_methods:
-            raise InvalidInputValue("read_method", valid_methods)
-        return v
+        if self.read_method not in valid_methods:
+            raise InputValueError("read_method", valid_methods)
+        self.get_service_options()
+        if self.validation:
+            self.validate_wfs()
 
-    @validator("crs")
-    def _valid_crs(cls, v: str) -> str:
-        return utils.validate_crs(v)
-
-    @validator("version")
-    def _version(cls, v: str) -> str:
-        return validate_version(v, ["1.0.0", "1.1.0", "2.0.0"])
-
-    def validate_wfs(self) -> None:
+    def get_service_options(self) -> None:
         """Validate input arguments with the WFS service."""
         try:
             wfs = WebFeatureService(self.url, version=self.version)
         except AttributeError as ex:
-            raise ServiceUnavailable(self.url) from ex
+            raise ServiceUnavailableError(self.url) from ex
 
-        valid_layers = list(wfs.contents)
-        if self.layer is None:
-            raise MissingInputs(
-                "The layer argument is missing."
-                + " The following layers are available:\n"
-                + ", ".join(valid_layers)
-            )
-        if self.layer not in valid_layers:
-            raise InvalidInputValue("layers", valid_layers)
+        self.available_layer = list(wfs.contents)
 
         wfs_features = wfs.getOperationByName("GetFeature")
         if self.version == "1.0.0":
-            valid_outformats = [f.rsplit("}", 1)[-1] for f in wfs_features.formatOptions]
+            self.available_outformat = [f.rsplit("}", 1)[-1] for f in wfs_features.formatOptions]
         else:
-            valid_outformats = wfs_features.parameters["outputFormat"]["values"]
+            self.available_outformat = wfs_features.parameters["outputFormat"]["values"]
+        self.available_outformat = [f.lower() for f in self.available_outformat]
 
-        valid_outformats = [v.lower() for v in valid_outformats]
-        if self.outformat is None:
-            raise MissingInputs(
-                "The outformat argument is missing."
-                + " The following output formats are available:\n"
-                + ", ".join(valid_outformats)
+        self.available_crs = [f"{s.authority.lower()}:{s.code}" for s in wfs[self.layer].crsOptions]
+
+    def validate_wfs(self) -> None:
+        """Validate input arguments with the WFS service."""
+        if self.layer is None:
+            raise MissingInputError(
+                "The layer argument is missing."
+                + " The following layers are available:\n"
+                + ", ".join(self.available_layer)
             )
+        if self.layer not in self.available_layer:
+            raise InputValueError("layers", self.available_layer)
 
-        if self.outformat not in valid_outformats:
-            raise InvalidInputValue("outformat", valid_outformats)
+        if self.outformat not in self.available_outformat:
+            raise InputValueError("outformat", self.available_outformat)
 
-        valid_crss = [f"{s.authority.lower()}:{s.code}" for s in wfs[self.layer].crsOptions]
-        if self.crs.lower() not in valid_crss:
-            raise InvalidInputValue("crs", valid_crss)
+        if self.crs_str.lower() not in self.available_crs:
+            raise InputValueError("crs", self.available_crs)
 
-    def get_validnames(self) -> List[str]:
+    def get_validnames(self) -> list[str]:
         """Get valid column names for a layer."""
         max_features = "count" if self.version == "2.0.0" else "maxFeatures"
 
@@ -685,7 +629,7 @@ class WFSBase(BaseModel):
                 f"Version: {self.version}",
                 f"Layer: {self.layer}",
                 f"Output Format: {self.outformat}",
-                f"Output CRS: {self.crs}",
+                f"Output CRS: {self.crs_str}",
                 f"Read Method: {self.read_method}",
             ]
         )
