@@ -1,8 +1,9 @@
 """Base classes and function for REST, WMS, and WMF services."""
+from __future__ import annotations
+
 import itertools
 import uuid
-from ssl import SSLContext
-from typing import Any, Dict, Iterator, List, NamedTuple, Optional, Sequence, Tuple, Union
+from typing import TYPE_CHECKING, Any, Iterator, NamedTuple, Sequence, Union
 
 import async_retriever as ar
 import cytoolz as tlz
@@ -11,8 +12,13 @@ import shapely.ops as ops
 from shapely.geometry import LineString, MultiPoint, MultiPolygon, Point, Polygon
 
 from . import utils
-from .core import DEF_CRS, ArcGISRESTfulBase, WFSBase, WMSBase
-from .exceptions import InvalidInputType, InvalidInputValue, ZeroMatched
+from .core import ArcGISRESTfulBase, WFSBase, WMSBase
+from .exceptions import InputTypeError, InputValueError, ZeroMatchedError
+
+if TYPE_CHECKING:
+    from ssl import SSLContext
+
+CRSTYPE = Union[int, str, pyproj.CRS]
 
 
 class ArcGISRESTful:
@@ -23,7 +29,7 @@ class ArcGISRESTful:
     By default, all retrieval methods retry to get the missing feature IDs,
     if there are any. You can disable this behavior by setting ``disable_retry``
     to ``True``. If there are any missing feature IDs after the retry,
-    they are saved to a text file, path of which can be accessed by
+    they are saved to a text file, ipath of which can be accessed by
     ``self.client.failed_path``.
 
     Parameters
@@ -40,7 +46,7 @@ class ArcGISRESTful:
     outfields : str or list
         The output fields to be requested. Setting ``*`` as outfields requests
         all the available fields which is the default behaviour.
-    crs : str, optional
+    crs : str, int, or pyproj.CRS, optional
         The spatial reference of the output data, defaults to ``epsg:4326``.
     max_workers : int, optional
         Number of simultaneous download, default to 1, i.e., no threading. Note
@@ -54,16 +60,16 @@ class ArcGISRESTful:
     disable_retry : bool, optional
         If ``True`` in case there are any failed queries, no retrying attempts
         is done and object IDs of the failed requests is saved to a text file
-        which its path can be accessed via ``self.client.failed_path``.
+        which its ipath can be accessed via ``self.client.failed_path``.
     """
 
     def __init__(
         self,
         base_url: str,
-        layer: Optional[int] = None,
+        layer: int | None = None,
         outformat: str = "geojson",
-        outfields: Union[List[str], str] = "*",
-        crs: str = DEF_CRS,
+        outfields: list[str] | str = "*",
+        crs: CRSTYPE = 4326,
         max_workers: int = 1,
         verbose: bool = False,
         disable_retry: bool = False,
@@ -81,20 +87,20 @@ class ArcGISRESTful:
 
     def oids_bygeom(
         self,
-        geom: Union[
-            LineString,
-            Polygon,
-            Point,
-            MultiPoint,
-            Tuple[float, float],
-            List[Tuple[float, float]],
-            Tuple[float, float, float, float],
-        ],
-        geo_crs: Union[str, pyproj.CRS] = DEF_CRS,
+        geom: (
+            LineString
+            | Polygon
+            | Point
+            | MultiPoint
+            | tuple[float, float]
+            | list[tuple[float, float]]
+            | tuple[float, float, float, float]
+        ),
+        geo_crs: CRSTYPE = 4326,
         spatial_relation: str = "esriSpatialRelIntersects",
-        sql_clause: Optional[str] = None,
-        distance: Optional[int] = None,
-    ) -> Iterator[Tuple[str, ...]]:
+        sql_clause: str | None = None,
+        distance: int | None = None,
+    ) -> Iterator[tuple[str, ...]]:
         """Get feature IDs within a geometry that can be combined with a SQL where clause.
 
         Parameters
@@ -103,8 +109,8 @@ class ArcGISRESTful:
             A geometry (LineString, Polygon, Point, MultiPoint), tuple of length two
             (``(x, y)``), a list of tuples of length 2 (``[(x, y), ...]``), or bounding box
             (tuple of length 4 (``(xmin, ymin, xmax, ymax)``)).
-        geo_crs : str or pyproj.CRS
-            The spatial reference of the input geometry.
+        geo_crs : str, int, or pyproj.CRS, optional
+            The spatial reference of the input geometry, defaults to ``epsg:4326``.
         spatial_relation : str, optional
             The spatial relationship to be applied on the input geometry
             while performing the query. If not correct a list of available options is shown.
@@ -142,7 +148,7 @@ class ArcGISRESTful:
             "esriSpatialRelRelation",
         ]
         if spatial_relation not in valid_spatialrels:
-            raise InvalidInputValue("spatial_relation", valid_spatialrels)
+            raise InputValueError("spatial_relation", valid_spatialrels)
 
         if isinstance(geom, tuple) and len(geom) == 2:
             geom = Point(geom)
@@ -171,9 +177,9 @@ class ArcGISRESTful:
             return self.partition_oids(resp["objectIds"])
         except (KeyError, TypeError) as ex:
             msg = resp["error"]["message"] if "error" in resp else "No matched records"
-            raise ZeroMatched(msg) from ex
+            raise ZeroMatchedError(msg) from ex
 
-    def oids_byfield(self, field: str, ids: Union[str, List[str]]) -> Iterator[Tuple[str, ...]]:
+    def oids_byfield(self, field: str, ids: str | list[str]) -> Iterator[tuple[str, ...]]:
         """Get Object IDs based on a list of field IDs.
 
         Parameters
@@ -189,10 +195,10 @@ class ArcGISRESTful:
             A list of feature IDs partitioned by ``self.max_nrecords``.
         """
         if field not in self.client.valid_fields:
-            raise InvalidInputValue("field", self.client.valid_fields)
+            raise InputValueError("field", self.client.valid_fields)
 
         if not isinstance(ids, Sequence):
-            raise InvalidInputType("ids", "str or list")
+            raise InputTypeError("ids", "str or list")
 
         ids_ls = [ids] if isinstance(ids, str) else ids
 
@@ -204,7 +210,7 @@ class ArcGISRESTful:
 
         return self.oids_bysql(f"{field} IN ({fids})")
 
-    def oids_bysql(self, sql_clause: str) -> Iterator[Tuple[str, ...]]:
+    def oids_bysql(self, sql_clause: str) -> Iterator[tuple[str, ...]]:
         """Get feature IDs using a valid SQL 92 WHERE clause.
 
         Notes
@@ -223,7 +229,7 @@ class ArcGISRESTful:
             A list of feature IDs partitioned by ``self.max_nrecords``.
         """
         if not isinstance(sql_clause, str):
-            raise InvalidInputType("sql_clause", "str")
+            raise InputTypeError("sql_clause", "str")
 
         payload = {
             "where": sql_clause,
@@ -238,9 +244,9 @@ class ArcGISRESTful:
             return self.partition_oids(resp["objectIds"])
         except (KeyError, TypeError) as ex:
             msg = resp["error"]["message"] if "error" in resp else "No matched records"
-            raise ZeroMatched(msg) from ex
+            raise ZeroMatchedError(msg) from ex
 
-    def partition_oids(self, oids: Union[List[int], int]) -> Iterator[Tuple[str, ...]]:
+    def partition_oids(self, oids: list[int] | int) -> Iterator[tuple[str, ...]]:
         """Partition feature IDs based on ``self.max_nrecords``.
 
         Parameters
@@ -257,10 +263,10 @@ class ArcGISRESTful:
 
     def get_features(
         self,
-        featureids: Iterator[Tuple[str, ...]],
+        featureids: Iterator[tuple[str, ...]],
         return_m: bool = False,
         return_geom: bool = True,
-    ) -> List[Dict[str, Any]]:
+    ) -> list[dict[str, Any]]:
         """Get features based on the feature IDs.
 
         Parameters
@@ -298,7 +304,7 @@ class WMS:
     outformat : str
         The data format to request for data from the service. You can pass an empty
         string to get a list of available output formats.
-    crs : str, optional
+    crs : str, int, or pyproj.CRS, optional
         The spatial reference system to be used for requesting the data, defaults to
         ``epsg:4326``.
     version : str, optional
@@ -315,12 +321,12 @@ class WMS:
     def __init__(
         self,
         url: str,
-        layers: Union[str, List[str]],
+        layers: str | list[str],
         outformat: str,
         version: str = "1.3.0",
-        crs: str = DEF_CRS,
+        crs: CRSTYPE = 4326,
         validation: bool = True,
-        ssl: Union[SSLContext, bool, None] = None,
+        ssl: SSLContext | bool | None = None,
     ) -> None:
         self.client = WMSBase(
             url=url,
@@ -328,31 +334,31 @@ class WMS:
             outformat=outformat,
             version=version,
             crs=crs,
+            validation=validation,
         )
         self.url = self.client.url
         self.outformat = self.client.outformat
         self.version = self.client.version
         self.crs = self.client.crs
+        self.crs_str = self.client.crs_str
         self.ssl = ssl
         self.layers = (
             [self.client.layers] if isinstance(self.client.layers, str) else self.client.layers
         )
-        if validation:
-            self.client.validate_wms()
 
-    def get_validlayers(self) -> Dict[str, str]:
+    def get_validlayers(self) -> dict[str, str]:
         """Get the layers supported by the WMS service."""
         return self.client.get_validlayers()
 
     def getmap_bybox(
         self,
-        bbox: Tuple[float, float, float, float],
+        bbox: tuple[float, float, float, float],
         resolution: float,
-        box_crs: Union[str, pyproj.CRS] = DEF_CRS,
+        box_crs: CRSTYPE = 4326,
         always_xy: bool = False,
         max_px: int = 8000000,
-        kwargs: Optional[Dict[str, Any]] = None,
-    ) -> Dict[str, bytes]:
+        kwargs: dict[str, Any] | None = None,
+    ) -> dict[str, bytes]:
         """Get data from a WMS service within a geometry or bounding box.
 
         Parameters
@@ -362,7 +368,7 @@ class WMS:
         resolution : float
             The output resolution in meters. The width and height of output are computed in pixel
             based on the geometry bounds and the given resolution.
-        box_crs : str, or pyproj.CRS, optional
+        box_crs : str, int, or pyproj.CRS, optional
             The spatial reference system of the input bbox, defaults to
             ``epsg:4326``.
         always_xy : bool, optional
@@ -384,8 +390,8 @@ class WMS:
             from the WMS service as bytes.
         """
         utils.check_bbox(bbox)
-        _bbox = utils.match_crs(bbox, box_crs, self.crs)
-        bounds = utils.bbox_decompose(_bbox, resolution, self.crs, max_px)
+        _bbox = utils.match_crs(bbox, box_crs, self.crs_str)
+        bounds = utils.bbox_decompose(_bbox, resolution, self.crs_str, max_px)
 
         payload = {
             "version": self.version,
@@ -394,23 +400,23 @@ class WMS:
         }
 
         if not isinstance(kwargs, (dict, type(None))):
-            raise InvalidInputType("kwargs", "dict or None")
+            raise InputTypeError("kwargs", "dict or None")
 
         if isinstance(kwargs, dict):
             payload.update(kwargs)
 
         if self.version == "1.1.1":
-            payload["srs"] = self.crs
+            payload["srs"] = self.crs_str
         else:
-            payload["crs"] = self.crs
+            payload["crs"] = self.crs_str
 
         def _get_payloads(
-            args: Tuple[str, Tuple[Tuple[float, float, float, float], str, int, int]]
-        ) -> Tuple[str, Dict[str, str]]:
+            args: tuple[str, tuple[tuple[float, float, float, float], str, int, int]]
+        ) -> tuple[str, dict[str, str]]:
             lyr, bnds = args
             _bbox, counter, _width, _height = bnds
 
-            if self.version != "1.1.1" and pyproj.CRS(self.crs).is_geographic and not always_xy:
+            if self.version != "1.1.1" and pyproj.CRS(self.crs_str).is_geographic and not always_xy:
                 _bbox = (_bbox[1], _bbox[0], _bbox[3], _bbox[2])
             _payload = payload.copy()
             _payload["bbox"] = f'{",".join(str(c) for c in _bbox)}'
@@ -450,7 +456,7 @@ class WFS(WFSBase):
     version : str, optional
         The WFS service version which should be either 1.0.0, 1.1.0, or 2.0.0.
         Defaults to 2.0.0.
-    crs: str, optional
+    crs : str, int, or pyproj.CRS, optional
         The spatial reference system to be used for requesting the data, defaults to
         ``epsg:4326``.
     read_method : str, optional
@@ -469,10 +475,10 @@ class WFS(WFSBase):
     def __init__(
         self,
         url: str,
-        layer: Optional[str] = None,
-        outformat: Optional[str] = None,
+        layer: str | None = None,
+        outformat: str | None = None,
         version: str = "2.0.0",
-        crs: str = DEF_CRS,
+        crs: CRSTYPE = 4326,
         read_method: str = "json",
         max_nrecords: int = 1000,
         validation: bool = True,
@@ -485,17 +491,15 @@ class WFS(WFSBase):
             crs=crs,
             read_method=read_method,
             max_nrecords=max_nrecords,
+            validation=validation,
         )
 
-        if validation:
-            self.validate_wfs()
-
     def getfeature_bybox(
         self,
-        bbox: Tuple[float, float, float, float],
-        box_crs: Union[str, pyproj.CRS] = DEF_CRS,
+        bbox: tuple[float, float, float, float],
+        box_crs: CRSTYPE = 4326,
         always_xy: bool = False,
-    ) -> Union[str, bytes, Dict[str, Any]]:
+    ) -> str | bytes | dict[str, Any]:
         """Get data from a WFS service within a bounding box.
 
         Parameters
@@ -529,18 +533,18 @@ class WFS(WFSBase):
             "request": "GetFeature",
             "typeName": self.layer,
             "bbox": f'{",".join(str(c) for c in bbox)},{box_crs.to_string()}',
-            "srsName": self.crs,
+            "srsName": self.crs_str,
         }
 
         return ar.retrieve([self.url], self.read_method, [{"params": payload}])[0]
 
     def getfeature_bygeom(
         self,
-        geometry: Union[Polygon, MultiPolygon],
-        geo_crs: Union[str, pyproj.CRS] = DEF_CRS,
+        geometry: Polygon | MultiPolygon,
+        geo_crs: CRSTYPE = 4326,
         always_xy: bool = False,
         predicate: str = "INTERSECTS",
-    ) -> Union[str, bytes, Dict[str, Any]]:
+    ) -> str | bytes | dict[str, Any]:
         """Get features based on a geometry.
 
         Parameters
@@ -574,7 +578,7 @@ class WFS(WFSBase):
         str or bytes or dict
             WFS query response based on the given geometry.
         """
-        geom = utils.match_crs(geometry, geo_crs, self.crs)
+        geom = utils.match_crs(geometry, geo_crs, self.crs_str)
 
         if self.version != "1.0.0" and pyproj.CRS(geo_crs).is_geographic and not always_xy:
             g_wkt = ops.transform(lambda x, y: (y, x), geom).wkt
@@ -594,15 +598,15 @@ class WFS(WFSBase):
             "BEYOND",
         ]
         if predicate not in valid_predicates:
-            raise InvalidInputValue("predicate", valid_predicates)
+            raise InputValueError("predicate", valid_predicates)
 
         return self.getfeature_byfilter(f"{predicate.upper()}(the_geom, {g_wkt})", method="POST")
 
     def getfeature_byid(
         self,
         featurename: str,
-        featureids: Union[List[str], str],
-    ) -> List[Union[str, bytes, Dict[str, Any]]]:
+        featureids: list[int | str] | int | str,
+    ) -> list[str | bytes | dict[str, Any]]:
         """Get features based on feature IDs.
 
         Parameters
@@ -619,15 +623,15 @@ class WFS(WFSBase):
         """
         valid_features = self.get_validnames()
         if featurename not in valid_features:
-            raise InvalidInputValue("featurename", valid_features)
+            raise InputValueError("featurename", valid_features)
 
-        if not isinstance(featureids, (str, int, list)):
-            raise InvalidInputType("featureids", "str or list of str")
+        if not isinstance(featureids, (str, int, list, tuple)):
+            raise InputTypeError("featureids", "str or list of str")
 
-        featureids = [featureids] if isinstance(featureids, (str, int)) else featureids
+        featureids = [featureids] if isinstance(featureids, (str, int)) else list(featureids)
 
         if len(featureids) == 0:
-            raise InvalidInputType("featureids", "int or str or list")
+            raise InputTypeError("featureids", "int or str or list")
 
         fid_list = (
             ", ".join(f"'{fid}'" for fid in fids)
@@ -641,7 +645,7 @@ class WFS(WFSBase):
 
     def getfeature_byfilter(
         self, cql_filter: str, method: str = "GET"
-    ) -> Union[str, bytes, Dict[str, Any]]:
+    ) -> str | bytes | dict[str, Any]:
         """Get features based on a valid CQL filter.
 
         Notes
@@ -663,11 +667,11 @@ class WFS(WFSBase):
             WFS query response
         """
         if not isinstance(cql_filter, str):
-            raise InvalidInputType("cql_filter", "str")
+            raise InputTypeError("cql_filter", "str")
 
         valid_methods = ["GET", "POST"]
         if method not in valid_methods:
-            raise InvalidInputValue("method", valid_methods)
+            raise InputValueError("method", valid_methods)
 
         payload = {
             "service": "wfs",
@@ -675,7 +679,7 @@ class WFS(WFSBase):
             "outputFormat": self.outformat,
             "request": "GetFeature",
             "typeName": self.layer,
-            "srsName": self.crs,
+            "srsName": self.crs_str,
             "cql_filter": cql_filter,
         }
 
