@@ -1,13 +1,17 @@
 """Some utilities for PyGeoOGC."""
+from __future__ import annotations
+
+import itertools
 import math
 import os
 from dataclasses import dataclass
 from pathlib import Path
-from typing import Any, Dict, List, Mapping, Optional, Tuple, TypeVar, Union
+from typing import Any, List, Mapping, Tuple, TypeVar, Union
 
 import async_retriever as ar
-import defusedxml.ElementTree as etree
+import defusedxml.ElementTree as ETree
 import pyproj
+import requests
 import shapely.geometry as sgeom
 import ujson as json
 from requests.adapters import HTTPAdapter
@@ -17,11 +21,10 @@ from requests_cache.backends.sqlite import SQLiteCache
 from shapely import ops
 from urllib3 import Retry
 
-from .exceptions import InvalidInputType, ServiceError
+from .exceptions import InputTypeError, ServiceError
 
-DEF_CRS = "epsg:4326"
+CRSTYPE = Union[int, str, pyproj.CRS]
 BOX_ORD = "(west, south, east, north)"
-EXPIRE = -1
 G = TypeVar(
     "G",
     sgeom.Point,
@@ -38,8 +41,8 @@ G = TypeVar(
 def check_response(resp: str) -> str:
     """Extract error message from a response, if any."""
     try:
-        root = etree.fromstring(resp)
-    except etree.ParseError:
+        root = ETree.fromstring(resp)
+    except ETree.ParseError:
         return resp
     else:
         try:
@@ -51,8 +54,10 @@ def check_response(resp: str) -> str:
 class RetrySession:
     """Configures the passed-in session to retry on failed requests.
 
+    Notes
+    -----
     The fails can be due to connection errors, specific HTTP response
-    codes and 30X redirections. The code is was originally based on:
+    codes and 30X redirections. The code was originally based on:
     https://github.com/bustawin/retry-requests
 
     Parameters
@@ -70,24 +75,32 @@ class RetrySession:
         system's temp directory.
     expire_after : int, optional
         Expiration time for the cache in seconds, defaults to -1 (never expire).
+    disable : bool, optional
+        If ``True`` temporarily disable caching requests and get new responses
+        from the server, defaults to ``False``.
     """
 
     def __init__(
         self,
         retries: int = 3,
         backoff_factor: float = 0.3,
-        status_to_retry: Tuple[int, ...] = (500, 502, 504),
-        prefixes: Tuple[str, ...] = ("https://",),
-        cache_name: Optional[Union[str, Path]] = None,
-        expire_after: int = EXPIRE,
+        status_to_retry: tuple[int, ...] = (500, 502, 504),
+        prefixes: tuple[str, ...] = ("https://",),
+        cache_name: str | Path | None = None,
+        expire_after: int = -1,
+        disable: bool = False,
     ) -> None:
-        self.cache_name = (
-            Path("cache", "http_cache.sqlite") if cache_name is None else Path(cache_name)
-        )
-        backend = SQLiteCache(self.cache_name, fast_save=True, timeout=1)
-        self.session = CachedSession(
-            expire_after=int(os.getenv("HYRIVER_CACHE_EXPIRE", expire_after)), backend=backend
-        )
+        disable = os.getenv("HYRIVER_CACHE_DISABLE", f"{disable}").lower() == "true"
+        if disable:
+            self.session = requests.Session()
+        else:
+            self.cache_name = (
+                Path("cache", "http_cache.sqlite") if cache_name is None else Path(cache_name)
+            )
+            backend = SQLiteCache(self.cache_name, fast_save=True, timeout=1)
+            self.session = CachedSession(
+                expire_after=int(os.getenv("HYRIVER_CACHE_EXPIRE", expire_after)), backend=backend
+            )
 
         adapter = HTTPAdapter(
             max_retries=Retry(
@@ -105,8 +118,8 @@ class RetrySession:
     def get(
         self,
         url: str,
-        payload: Optional[Mapping[str, Any]] = None,
-        headers: Optional[Mapping[str, Any]] = None,
+        payload: Mapping[str, Any] | None = None,
+        headers: Mapping[str, Any] | None = None,
     ) -> Response:
         """Retrieve data from a url by GET and return the Response."""
         resp = self.session.get(url, params=payload, headers=headers)
@@ -120,8 +133,8 @@ class RetrySession:
     def post(
         self,
         url: str,
-        payload: Optional[Mapping[str, Any]] = None,
-        headers: Optional[Mapping[str, Any]] = None,
+        payload: Mapping[str, Any] | None = None,
+        headers: Mapping[str, Any] | None = None,
     ) -> Response:
         """Retrieve data from a url by POST and return the Response."""
         resp = self.session.post(url, data=payload, headers=headers)
@@ -132,25 +145,41 @@ class RetrySession:
         else:
             return resp
 
+    def head(
+        self,
+        url: str,
+        data: Mapping[str, Any] | None = None,
+        params: Mapping[str, Any] | None = None,
+        headers: Mapping[str, Any] | None = None,
+    ) -> Response:
+        """Retrieve data from a url by POST and return the Response."""
+        resp = self.session.head(url, data=data, params=params, headers=headers)
+        try:
+            resp.raise_for_status()
+        except RequestException as ex:
+            raise ServiceError(check_response(resp.text)) from ex
+        else:
+            return resp
+
 
 def traverse_json(
-    obj: Union[Dict[str, Any], List[Dict[str, Any]]], path: Union[str, List[str]]
-) -> List[Any]:
-    """Extract an element from a JSON file along a specified path.
+    items: dict[str, Any] | list[dict[str, Any]], ipath: str | list[str]
+) -> list[Any]:
+    """Extract an element from a JSON file along a specified ipath.
 
     This function is based on `bcmullins <https://bcmullins.github.io/parsing-json-python/>`__.
 
     Parameters
     ----------
-    obj : dict
+    items : dict
         The input json dictionary
-    path : list
-        The path to the requested element
+    ipath : list
+        The ipath to the requested element
 
     Returns
     -------
     list
-        The items founds in the JSON
+        The sub_items founds in the JSON
 
     Examples
     --------
@@ -166,45 +195,45 @@ def traverse_json(
     """
 
     def extract(
-        obj: Optional[Union[List[Any], Dict[str, Any]]],
-        path: Union[str, List[str]],
+        sub_items: list[Any] | dict[str, Any] | None,
+        path: str | list[str],
         ind: int,
-        arr: List[Any],
-    ) -> List[Any]:
+        arr: list[Any],
+    ) -> list[Any]:
         key = path[ind]
         if ind + 1 < len(path):
-            if isinstance(obj, dict):
-                if key in obj:
-                    extract(obj.get(key), path, ind + 1, arr)
+            if isinstance(sub_items, dict):
+                if key in sub_items:
+                    extract(sub_items.get(key), path, ind + 1, arr)
                 else:
                     arr.append(None)
-            elif isinstance(obj, list):
-                if not obj:
+            elif isinstance(sub_items, list):
+                if not sub_items:
                     arr.append(None)
                 else:
-                    for item in obj:
-                        extract(item, path, ind, arr)
+                    for i in sub_items:
+                        extract(i, path, ind, arr)
             else:
                 arr.append(None)
         if ind + 1 == len(path):
-            if isinstance(obj, list):
-                if not obj:
+            if isinstance(sub_items, list):
+                if not sub_items:
                     arr.append(None)
                 else:
-                    for item in obj:
-                        arr.append(item.get(key))
-            elif isinstance(obj, dict):
-                arr.append(obj.get(key))
+                    for i in sub_items:
+                        arr.append(i.get(key))
+            elif isinstance(sub_items, dict):
+                arr.append(sub_items.get(key))
             else:
                 arr.append(None)
         return arr
 
-    if isinstance(obj, dict):
-        return extract(obj, path, 0, [])
+    if isinstance(items, dict):
+        return extract(items, ipath, 0, [])
 
     outer_arr = []
-    for item in obj:
-        outer_arr.append(extract(item, path, 0, []))
+    for item in items:
+        outer_arr.append(extract(item, ipath, 0, []))
     return outer_arr
 
 
@@ -224,15 +253,16 @@ class ESRIGeomQuery:
         for reference.
     """
 
-    geometry: Union[
-        Tuple[float, float],
-        List[Tuple[float, float]],
-        Tuple[float, float, float, float],
-        sgeom.Polygon,
-    ]
+    geometry: (
+        tuple[float, float]
+        | list[tuple[float, float]]
+        | tuple[float, float, float, float]
+        | sgeom.Polygon
+        | sgeom.LineString
+    )
     wkid: int
 
-    def _get_payload(self, geo_type: str, geo_json: Dict[str, Any]) -> Mapping[str, str]:
+    def _get_payload(self, geo_type: str, geo_json: dict[str, Any]) -> Mapping[str, str]:
         """Generate a request payload based on ESRI template.
 
         Parameters
@@ -257,7 +287,7 @@ class ESRIGeomQuery:
     def point(self) -> Mapping[str, str]:
         """Query for a point."""
         if not (isinstance(self.geometry, tuple) and len(self.geometry) == 2):
-            raise InvalidInputType("geometry", "tuple", "(x, y)")
+            raise InputTypeError("geometry", "tuple", "(x, y)")
 
         geo_type = "esriGeometryPoint"
         geo_json = dict(zip(("x", "y"), self.geometry))
@@ -266,7 +296,7 @@ class ESRIGeomQuery:
     def multipoint(self) -> Mapping[str, str]:
         """Query for a multi-point."""
         if not (isinstance(self.geometry, list) and all(len(g) == 2 for g in self.geometry)):
-            raise InvalidInputType("geometry", "list of tuples", "[(x, y), ...]")
+            raise InputTypeError("geometry", "list of tuples", "[(x, y), ...]")
 
         geo_type = "esriGeometryMultipoint"
         geo_json = {"points": [[x, y] for x, y in self.geometry]}
@@ -275,7 +305,7 @@ class ESRIGeomQuery:
     def bbox(self) -> Mapping[str, str]:
         """Query for a bbox."""
         if not (isinstance(self.geometry, (tuple, list)) and len(self.geometry) == 4):
-            raise InvalidInputType("geometry", "tuple", BOX_ORD)
+            raise InputTypeError("geometry", "tuple", BOX_ORD)
 
         geo_type = "esriGeometryEnvelope"
         geo_json = dict(zip(("xmin", "ymin", "xmax", "ymax"), self.geometry))
@@ -284,7 +314,7 @@ class ESRIGeomQuery:
     def polygon(self) -> Mapping[str, str]:
         """Query for a polygon."""
         if not isinstance(self.geometry, sgeom.Polygon):
-            raise InvalidInputType("geometry", "Polygon")
+            raise InputTypeError("geometry", "Polygon")
 
         geo_type = "esriGeometryPolygon"
         geo_json = {"rings": [[[x, y] for x, y in zip(*self.geometry.exterior.coords.xy)]]}
@@ -293,14 +323,14 @@ class ESRIGeomQuery:
     def polyline(self) -> Mapping[str, str]:
         """Query for a polyline."""
         if not isinstance(self.geometry, sgeom.LineString):
-            raise InvalidInputType("geometry", "LineString")
+            raise InputTypeError("geometry", "LineString")
 
         geo_type = "esriGeometryPolyline"
         geo_json = {"paths": [[[x, y] for x, y in zip(*self.geometry.coords.xy)]]}
         return self._get_payload(geo_type, geo_json)
 
 
-def match_crs(geom: G, in_crs: str, out_crs: str) -> G:
+def match_crs(geom: G, in_crs: CRSTYPE, out_crs: CRSTYPE) -> G:
     """Reproject a geometry to another CRS.
 
     Parameters
@@ -309,9 +339,9 @@ def match_crs(geom: G, in_crs: str, out_crs: str) -> G:
         Input geometry which could be a list of coordinates such as ``[(x1, y1), ...]``,
         a bounding box like so ``(xmin, ymin, xmax, ymax)``, or any valid ``shapely``'s
         geometry such as ``Polygon``, ``MultiPolygon``, etc..
-    in_crs : str
+    in_crs : str, int, or pyproj.CRS
         Spatial reference of the input geometry
-    out_crs : str
+    out_crs : str, int, or pyproj.CRS
         Target spatial reference
 
     Returns
@@ -333,6 +363,9 @@ def match_crs(geom: G, in_crs: str, out_crs: str) -> G:
     >>> match_crs(coords, "epsg:3857", "epsg:4326")
     [(-69.7636111130079, 45.44549114818127)]
     """
+    if pyproj.CRS(in_crs) == pyproj.CRS(out_crs):
+        return geom
+
     project = pyproj.Transformer.from_crs(in_crs, out_crs, always_xy=True).transform
 
     if isinstance(
@@ -359,55 +392,21 @@ def match_crs(geom: G, in_crs: str, out_crs: str) -> G:
         "a list of coordinates such as [(x1, y1), ...],"
         + "a bounding box like so (xmin, ymin, xmax, ymax), or any valid shapely's geometry."
     )
-    raise InvalidInputType("geom", gtypes)
+    raise InputTypeError("geom", gtypes)
 
 
-def check_bbox(bbox: Tuple[float, float, float, float]) -> None:
+def check_bbox(bbox: tuple[float, float, float, float]) -> None:
     """Check if an input inbox is a tuple of length 4."""
     if not (isinstance(bbox, tuple) and len(bbox) == 4):
-        raise InvalidInputType("bbox", "tuple", BOX_ORD)
-
-
-def bbox_resolution(
-    bbox: Tuple[float, float, float, float], resolution: float, bbox_crs: str = DEF_CRS
-) -> Tuple[int, int]:
-    """Image size of a bounding box WGS84 for a given resolution in meters.
-
-    Parameters
-    ----------
-    bbox : tuple
-        A bounding box in WGS84 (west, south, east, north)
-    resolution : float
-        The resolution in meters
-    bbox_crs : str, optional
-        The spatial reference of the input bbox, default to EPSG:4326.
-
-    Returns
-    -------
-    tuple
-        The width and height of the image
-    """
-    check_bbox(bbox)
-
-    bbox = match_crs(bbox, bbox_crs, DEF_CRS)
-    west, south, east, north = bbox
-    geod = pyproj.Geod(ellps="WGS84")
-
-    linex = sgeom.LineString([sgeom.Point(west, south), sgeom.Point(east, south)])
-    delx = geod.geometry_length(linex)
-
-    liney = sgeom.LineString([sgeom.Point(west, south), sgeom.Point(west, north)])
-    dely = geod.geometry_length(liney)
-
-    return int(delx / resolution), int(dely / resolution)
+        raise InputTypeError("bbox", "tuple", BOX_ORD)
 
 
 def bbox_decompose(
-    bbox: Tuple[float, float, float, float],
+    bbox: tuple[float, float, float, float],
     resolution: float,
-    box_crs: str = DEF_CRS,
+    box_crs: CRSTYPE = 4326,
     max_px: int = 8000000,
-) -> List[Tuple[Tuple[float, float, float, float], str, int, int]]:
+) -> list[tuple[tuple[float, float, float, float], str, int, int]]:
     r"""Split the bounding box vertically for WMS requests.
 
     Parameters
@@ -416,8 +415,8 @@ def bbox_decompose(
         A bounding box; (west, south, east, north)
     resolution : float
         The target resolution for a WMS request in meters.
-    box_crs : str, optional
-        The spatial reference of the input bbox, default to EPSG:4326.
+    box_crs : str, int, or pyproj.CRS, optional
+        The spatial reference of the input bbox, default to ``epsg:4326``.
     max_px : int, opitonal
         The maximum allowable number of pixels (width x height) for a WMS requests,
         defaults to 8 million based on some trial-and-error.
@@ -427,7 +426,7 @@ def bbox_decompose(
     list of tuples
         Each tuple includes the following elements:
 
-        * Tuple of length 4 that represents a bounding box (west, south, east, north) of a cell,
+        * Tuple of px_tot 4 that represents a bounding box (west, south, east, north) of a cell,
         * A label that represents cell ID starting from bottom-left to top-right, for example a
           2x2 decomposition has the following labels::
 
@@ -446,66 +445,48 @@ def bbox_decompose(
 
     """
     check_bbox(bbox)
-    _bbox = match_crs(bbox, box_crs, DEF_CRS)
-    width, height = bbox_resolution(_bbox, resolution, DEF_CRS)
 
-    n_px = width * height
-    if n_px < max_px:
-        return [(bbox, "0_0", width, height)]
+    geod = pyproj.Geod(ellps="GRS80")
 
-    geod = pyproj.Geod(ellps="WGS84")
-    west, south, east, north = _bbox
+    west, south, east, north = bbox
 
-    def directional_split(
-        az: float, origin: float, dest: float, lvl: float, xy: bool, px: int
-    ) -> Tuple[List[Tuple[float, Any]], List[int]]:
-        divs = [0]
-        mul = 1.0
-        coords = []
+    xmin, ymin, xmax, ymax = match_crs(bbox, box_crs, 4326)
 
-        def get_args(dst: float, dx: float) -> Tuple[float, float, float, float, int]:
-            return (dst, lvl, az, dx, 0) if xy else (lvl, dst, az, dx, 1)
+    x_dist = geod.geometry_length(sgeom.LineString([(xmin, ymin), (xmax, ymin)]))
+    y_dist = geod.geometry_length(sgeom.LineString([(xmin, ymin), (xmin, ymax)]))
+    width = int(math.ceil(x_dist / resolution))
+    height = int(math.ceil(y_dist / resolution))
 
-        while divs[-1] < 1:
-            dim = int(math.sqrt(max_px) * mul)
-            step = (dim - 1) * resolution
+    if width * height <= max_px:
+        bboxs = [(bbox, "0_0", width, height)]
 
-            _dest = origin
-            while _dest < dest:
-                args = get_args(_dest, step)
-                coords.append((_dest, geod.fwd(*args[:-1])[args[-1]]))
+    n_px = int(math.sqrt(max_px))
 
-                args = get_args(coords[-1][-1], resolution)
-                _dest = geod.fwd(*args[:-1])[args[-1]]
+    def _split_directional(low: float, high: float, px_tot: int) -> tuple[list[int], list[float]]:
+        npt = [n_px for _ in range(int(px_tot / n_px))] + [px_tot % n_px]
+        xd = abs(high - low)
+        dx = [xd * n / sum(npt) for n in npt]
+        xs = [low + d for d in itertools.accumulate(dx)]
+        xs.insert(0, low)
+        return npt, xs
 
-            coords[-1] = (coords[-1][0], dest)
-
-            n_dim = len(coords)
-            divs = [dim for _ in range(n_dim)]
-            divs[-1] = px - (n_dim - 1) * dim
-            mul -= 0.1
-        return coords, divs
-
-    az_x = geod.inv(west, south, east, south)[0]
-    lons, widths = directional_split(az_x, west, east, south, True, width)
-
-    az_y = geod.inv(west, south, west, north)[0]
-    lats, heights = directional_split(az_y, south, north, west, False, height)
+    nw, xs = _split_directional(west, east, width)
+    nh, ys = _split_directional(south, north, height)
 
     bboxs = []
-    for i, ((bottom, top), h) in enumerate(zip(lats, heights)):
-        for j, ((left, right), w) in enumerate(zip(lons, widths)):
-            bx_crs = match_crs((left, bottom, right, top), DEF_CRS, box_crs)
-            bboxs.append((bx_crs, f"{i}_{j}", w, h))
+    for j in range(len(nh)):
+        for i in range(len(nw)):
+            bx_crs = (xs[i], ys[j], xs[i + 1], ys[j + 1])
+            bboxs.append((bx_crs, f"{i}_{j}", nw[i], nh[j]))
     return bboxs
 
 
-def validate_crs(val: Union[str, int, pyproj.CRS]) -> str:
+def validate_crs(crs: CRSTYPE) -> str:
     """Validate a CRS.
 
     Parameters
     ----------
-    val : str or int
+    crs : str, int, or pyproj.CRS
         Input CRS.
 
     Returns
@@ -514,19 +495,18 @@ def validate_crs(val: Union[str, int, pyproj.CRS]) -> str:
         Validated CRS as a string.
     """
     try:
-        crs: str = pyproj.CRS(val).to_string()
+        return pyproj.CRS(crs).to_string()  # type: ignore
     except pyproj.exceptions.CRSError as ex:
-        raise InvalidInputType("crs", "a valid CRS") from ex
-    return crs
+        raise InputTypeError("crs", "a valid CRS") from ex
 
 
-def valid_wms_crs(url: str) -> List[str]:
+def valid_wms_crs(url: str) -> list[str]:
     """Get valid CRSs from a WMS service version 1.3.0."""
     ns = "http://www.opengis.net/wms"
 
-    def get_path(tag_list: List[str]) -> str:
+    def get_path(tag_list: list[str]) -> str:
         return f"/{{{ns}}}".join([""] + tag_list)[1:]
 
     kwds = {"params": {"service": "wms", "request": "GetCapabilities"}}
-    root = etree.fromstring(ar.retrieve_text([url], [kwds], ssl=False)[0])
+    root = ETree.fromstring(ar.retrieve_text([url], [kwds], ssl=False)[0])
     return [t.text.lower() for t in root.findall(get_path(["Capability", "Layer", "CRS"]))]
