@@ -124,6 +124,13 @@ class TestREST:
         assert len(resp[0]["features"]) == 3
 
 
+def test_retrysession_head():
+    url = "https://httpbin.org/image/jpeg"
+    session = utils.RetrySession()
+    resp = session.head(url)
+    assert resp.headers["Content-length"] == "35588"
+
+
 @pytest.mark.filterwarnings("ignore:.*Content metadata*.")
 class TestWMS:
     wms_url: str = ServiceURL().wms.gebco
@@ -142,7 +149,7 @@ class TestWMS:
         r_dict = wms.getmap_bybox(GEO_NAT.bounds, 20, DEF_CRS)
         assert (
             wms.get_validlayers()[self.layer] == self.layer
-            and sys.getsizeof(r_dict[f"{self.layer}_dd_0_0"]) == 11485083
+            and sys.getsizeof(r_dict[f"{self.layer}_dd_0_0"]) == 11496857
         )
 
     def test_bybox(self):
@@ -150,7 +157,7 @@ class TestWMS:
         wms = WMS(self.wms_url, layers=self.layer, outformat="image/tiff", crs=DEF_CRS)
         print(wms)
         r_dict = wms.getmap_bybox(GEO_NAT.bounds, 20, DEF_CRS, max_px=int(3e6))
-        assert sum(sys.getsizeof(r) for r in r_dict.values()) == 11495526
+        assert sum(sys.getsizeof(r) for r in r_dict.values()) == 11507302
 
     def test_valid_crs(self):
         """Get WMS valid CRSs"""
@@ -203,7 +210,7 @@ class TestWFS:
         r = self.wfs.getfeature_byfilter("staid LIKE '010315%'")
         assert self.to_df(r).shape[0] == 2
 
-    def test_wfs111(self):
+    def test_wfs100(self):
         """WFS 1.0.0 by geom"""
         wfs = WFS(
             ServiceURL().wfs.waterdata,
